@@ -1,7 +1,7 @@
-/* $Id: openssdpsocket.c,v 1.12 2012/05/21 17:13:11 nanard Exp $ */
+/* $Id: openssdpsocket.c,v 1.17 2015/08/06 14:05:37 nanard Exp $ */
 /* MiniUPnP project
  * http://miniupnp.free.fr/ or http://miniupnp.tuxfamily.org/
- * (c) 2006-2012 Thomas Bernard
+ * (c) 2006-2015 Thomas Bernard
  * This software is subject to the conditions detailed
  * in the LICENCE file provided within the distribution */
 
@@ -18,6 +18,9 @@
 
 #include "openssdpsocket.h"
 #include "upnputils.h"
+#include "minissdpdtypes.h"
+
+extern struct lan_addr_list lan_addrs;
 
 /* SSDP ip/port */
 #define SSDP_PORT (1900)
@@ -26,48 +29,6 @@
 #define LL_SSDP_MCAST_ADDR ("FF02::C")
 #define SL_SSDP_MCAST_ADDR ("FF05::C")
 
-/**
- * Get the IPv4 address from a string
- * representing the address or the interface name
- */
-static in_addr_t
-GetIfAddrIPv4(const char * ifaddr)
-{
-	in_addr_t addr;
-	int s;
-	struct ifreq ifr;
-	int ifrlen;
-
-	/* let's suppose ifaddr is a IPv4 address
-	 * such as 192.168.1.1 */
-	addr = inet_addr(ifaddr);
-	if(addr != INADDR_NONE)
-	{
-		return addr;
-	}
-	/* let's suppose the ifaddr was in fact an interface name
-	 * such as eth0 */
-	s = socket(PF_INET, SOCK_DGRAM, 0);
-	if(s < 0)
-	{
-		syslog(LOG_ERR, "socket(PF_INET, SOCK_DGRAM): %m");
-		return INADDR_NONE;
-	}
-	memset(&ifr, 0, sizeof(struct ifreq));
-	strncpy(ifr.ifr_name, ifaddr, IFNAMSIZ);
-	if(ioctl(s, SIOCGIFADDR, &ifr, &ifrlen) < 0)
-	{
-		syslog(LOG_ERR, "ioctl(s, SIOCGIFADDR, ...): %m");
-		close(s);
-		return INADDR_NONE;
-	}
-	syslog(LOG_DEBUG, "GetIfAddrIPv4(%s) = %s", ifaddr,
-	       inet_ntoa(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr));
-	addr = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr;
-	close(s);
-	return addr;
-}
-
 /**
  * Add the multicast membership for SSDP on the interface
  * @param s	the socket
@@ -75,30 +36,30 @@ GetIfAddrIPv4(const char * ifaddr)
  * @param ipv6	IPv6 or IPv4
  * return -1 on error, 0 on success */
 int
-AddDropMulticastMembership(int s, const char * ifaddr, int ipv6, int drop)
+AddDropMulticastMembership(int s, struct lan_addr_s * lan_addr, int ipv6, int drop)
 {
 	struct ip_mreq imr;	/* Ip multicast membership */
 #ifdef ENABLE_IPV6
 	struct ipv6_mreq mr;
-	unsigned int ifindex;
-#endif
+#else	/* ENABLE_IPV6 */
+	(void)ipv6;
+#endif	/* ENABLE_IPV6 */
 
 	if(s <= 0)
 		return -1;	/* nothing to do */
 #ifdef ENABLE_IPV6
 	if(ipv6)
 	{
-		ifindex = if_nametoindex(ifaddr);
 		memset(&mr, 0, sizeof(mr));
 		inet_pton(AF_INET6, LL_SSDP_MCAST_ADDR, &mr.ipv6mr_multiaddr);
-		mr.ipv6mr_interface = ifindex;
+		mr.ipv6mr_interface = lan_addr->index;
 		if(setsockopt(s, IPPROTO_IPV6, drop ? IPV6_LEAVE_GROUP : IPV6_JOIN_GROUP,
 		   &mr, sizeof(struct ipv6_mreq)) < 0)
 		{
 			syslog(LOG_ERR, "setsockopt(udp, %s)(%s, %s): %m",
 			       drop ? "IPV6_LEAVE_GROUP" : "IPV6_JOIN_GROUP",
 			       LL_SSDP_MCAST_ADDR,
-			       ifaddr);
+			       lan_addr->ifname);
 			return -1;
 		}
 		inet_pton(AF_INET6, SL_SSDP_MCAST_ADDR, &mr.ipv6mr_multiaddr);
@@ -108,59 +69,64 @@ AddDropMulticastMembership(int s, const char * ifaddr, int ipv6, int drop)
 			syslog(LOG_ERR, "setsockopt(udp, %s)(%s, %s): %m",
 			       drop ? "IPV6_LEAVE_GROUP" : "IPV6_JOIN_GROUP",
 			       SL_SSDP_MCAST_ADDR,
-			       ifaddr);
+			       lan_addr->ifname);
 			return -1;
 		}
 	}
 	else
 	{
-#endif
-	    /* setting up imr structure */
-	    imr.imr_multiaddr.s_addr = inet_addr(SSDP_MCAST_ADDR);
-	    /*imr.imr_interface.s_addr = htonl(INADDR_ANY);*/
-	    /*imr.imr_interface.s_addr = inet_addr(ifaddr);*/
-		imr.imr_interface.s_addr = GetIfAddrIPv4(ifaddr);
+#endif /* ENABLE_IPV6 */
+		/* setting up imr structure */
+		imr.imr_multiaddr.s_addr = inet_addr(SSDP_MCAST_ADDR);
+		imr.imr_interface.s_addr = lan_addr->addr.s_addr;
 		if(imr.imr_interface.s_addr == INADDR_NONE)
 		{
 			syslog(LOG_ERR, "no IPv4 address for interface %s",
-			       ifaddr);
+			       lan_addr->ifname);
 			return -1;
 		}
 
 		if (setsockopt(s, IPPROTO_IP, drop ? IP_DROP_MEMBERSHIP : IP_ADD_MEMBERSHIP,
 		    (void *)&imr, sizeof(struct ip_mreq)) < 0)
 		{
-	        syslog(LOG_ERR, "setsockopt(udp, %s)(%s): %m",
+			syslog(LOG_ERR, "setsockopt(udp, %s)(%s): %m",
 			       drop ? "IP_DROP_MEMBERSHIP" : "IP_ADD_MEMBERSHIP",
-			       ifaddr);
+			       lan_addr->ifname);
 			return -1;
-	    }
+		}
 #ifdef ENABLE_IPV6
 	}
-#endif
+#endif /* ENABLE_IPV6 */
 
 	return 0;
 }
 
 int
-OpenAndConfSSDPReceiveSocket(int n_listen_addr,
-							 const char * * listen_addr,
-                             int ipv6)
+OpenAndConfSSDPReceiveSocket(int ipv6, unsigned char ttl)
 {
 	int s;
 	int opt = 1;
+	unsigned char loopchar = 0;
 #ifdef ENABLE_IPV6
 	struct sockaddr_storage sockname;
-#else
+#else /* ENABLE_IPV6 */
 	struct sockaddr_in sockname;
-#endif
+#endif /* ENABLE_IPV6 */
 	socklen_t sockname_len;
+	struct lan_addr_s * lan_addr;
+
+#ifndef ENABLE_IPV6
+	if(ipv6) {
+		syslog(LOG_ERR, "%s: please compile with ENABLE_IPV6 to allow ipv6=1", __func__);
+		return -1;
+	}
+#endif /* ENABLE_IPV6 */
 
 #ifdef ENABLE_IPV6
 	if( (s = socket(ipv6 ? PF_INET6 : PF_INET, SOCK_DGRAM, 0)) < 0)
-#else
+#else /* ENABLE_IPV6 */
 	if( (s = socket(PF_INET, SOCK_DGRAM, 0)) < 0)
-#endif
+#endif /* ENABLE_IPV6 */
 	{
 		syslog(LOG_ERR, "socket(udp): %m");
 		return -1;
@@ -180,31 +146,66 @@ OpenAndConfSSDPReceiveSocket(int n_listen_addr,
 		{
 			syslog(LOG_WARNING, "setsockopt(IPV6_V6ONLY): %m");
 		}
-#endif
+#endif /* IPV6_V6ONLY */
 		struct sockaddr_in6 * sa = (struct sockaddr_in6 *)&sockname;
-    	sa->sin6_family = AF_INET6;
-    	sa->sin6_port = htons(SSDP_PORT);
+		sa->sin6_family = AF_INET6;
+		sa->sin6_port = htons(SSDP_PORT);
 		sa->sin6_addr = in6addr_any;
 		sockname_len = sizeof(struct sockaddr_in6);
 	}
 	else
 	{
 		struct sockaddr_in * sa = (struct sockaddr_in *)&sockname;
-    	sa->sin_family = AF_INET;
-    	sa->sin_port = htons(SSDP_PORT);
-    	sa->sin_addr.s_addr = htonl(INADDR_ANY);
+		sa->sin_family = AF_INET;
+		sa->sin_port = htons(SSDP_PORT);
+#ifdef SSDP_LISTEN_ON_SPECIFIC_ADDR
+		if(n_listen_addr == 1)
+		{
+			sa->sin_addr.s_addr = GetIfAddrIPv4(listen_addr[0]);
+			if(sa->sin_addr.s_addr == INADDR_NONE)
+			{
+				syslog(LOG_ERR, "no IPv4 address for interface %s",
+				       listen_addr[0]);
+				close(s);
+				return -1;
+			}
+		}
+		else
+#endif /* SSDP_LISTEN_ON_SPECIFIC_ADDR */
+			sa->sin_addr.s_addr = htonl(INADDR_ANY);
 		sockname_len = sizeof(struct sockaddr_in);
 	}
-#else
+#else /* ENABLE_IPV6 */
 	memset(&sockname, 0, sizeof(struct sockaddr_in));
     sockname.sin_family = AF_INET;
     sockname.sin_port = htons(SSDP_PORT);
-	/* NOTE : it seems it doesnt work when binding on the specific address */
-    /*sockname.sin_addr.s_addr = inet_addr(UPNP_MCAST_ADDR);*/
-    sockname.sin_addr.s_addr = htonl(INADDR_ANY);
-    /*sockname.sin_addr.s_addr = inet_addr(ifaddr);*/
+#ifdef SSDP_LISTEN_ON_SPECIFIC_ADDR
+	if(n_listen_addr == 1)
+	{
+		sockname.sin_addr.s_addr = GetIfAddrIPv4(listen_addr[0]);
+		if(sockname.sin_addr.s_addr == INADDR_NONE)
+		{
+			syslog(LOG_ERR, "no IPv4 address for interface %s",
+			       listen_addr[0]);
+			close(s);
+			return -1;
+		}
+	}
+	else
+#endif /* SSDP_LISTEN_ON_SPECIFIC_ADDR */
+	sockname.sin_addr.s_addr = htonl(INADDR_ANY);
 	sockname_len = sizeof(struct sockaddr_in);
-#endif
+#endif /* ENABLE_IPV6 */
+
+	if(setsockopt(s, IPPROTO_IP, IP_MULTICAST_LOOP, (char *)&loopchar, sizeof(loopchar)) < 0)
+	{
+		syslog(LOG_WARNING, "setsockopt(IP_MULTICAST_LOOP): %m");
+	}
+
+	if(setsockopt(s, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0)
+	{
+		syslog(LOG_WARNING, "setsockopt(IP_MULTICAST_TTL): %m");
+	}
 
 	if(setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) < 0)
 	{
@@ -218,14 +219,13 @@ OpenAndConfSSDPReceiveSocket(int n_listen_addr,
 		return -1;
     }
 
-	while(n_listen_addr>0)
+	for(lan_addr = lan_addrs.lh_first; lan_addr != NULL; lan_addr = lan_addr->list.le_next)
 	{
-		n_listen_addr--;
-		if(AddDropMulticastMembership(s, listen_addr[n_listen_addr], ipv6, 0) < 0)
+		if(AddDropMulticastMembership(s, lan_addr, ipv6, 0) < 0)
 		{
 			syslog(LOG_WARNING, "Failed to add IPv%d multicast membership for interface %s.",
 			       ipv6 ? 6 : 4,
-			       listen_addr[n_listen_addr] );
+			       lan_addr->ifname);
 		}
 	}
 
